from mospy import Transaction
from mospy.clients import HTTPClient

from src.mospy import Account

# As the account query returns a different json than the standard cosmos one evmos won't be compatible with the
# client.load_account() function
account = Account(
            seed_phrase="law grab theory better athlete submit awkward hawk state wedding wave monkey audit blame fury wood tag rent furnace exotic jeans drift destroy style",
            hrp="evmos",
            slip44=60,
            eth=True,
            next_sequence=1,
            account_number=2154050,
        )

tx = Transaction(
    account=account,
    gas=2000000,
    memo="The first mospy evmos transaction!",
    chain_id="evmos_9001-2",
)

tx.set_fee(
    denom="aevmos",
    amount=40000000000000000
)
tx.add_msg(
    tx_type="transfer",
    sender=account,
    receipient=account.address,
    amount=3500000000000000,
    denom="aevmos",
)

client = HTTPClient(api="https://api.evmos.interbloc.org")
tx_response = client.broadcast_transaction(transaction=tx)


from cosmospy_protobuf.cosmos.base.v1beta1.coin_pb2 import Coin
from cosmospy_protobuf.cosmos.staking.v1beta1.tx_pb2 import MsgDelegate

from src.mospy import Transaction, Account
from src.mospy.clients import HTTPClient

client = HTTPClient()
account = Account(
    seed_phrase="SEED HERE",
    account_number=1234,
    next_sequence=1234
)
tx = Transaction(
    account=account,
    gas=10000000  # This need to be adapted
)

dmsg = MsgDelegate(
        delegator_address=account.address,
        validator_address="VALOPER HERE",
        amount=Coin(amount=str(1000000), denom="uatom")
    )

tx.set_fee(
    amount=2000,
    denom="uatom"
)


tx.add_raw_msg(dmsg, type_url="/cosmos.staking.v1beta1.tx.MsgDelegate")

client.broadcast_transaction(transaction=tx)


import copy

import httpx
from mospy.Account import Account
from mospy.Transaction import Transaction

from mospy.exceptions.clients import NodeException


class HTTPClient:
    """
    Wrapper class to interact with a cosmos chain through their API endpoint

    Args:
        api (str): URL to a Api node
    """

    def __init__(self, *, api: str = "https://api.cosmos.interbloc.org"):
        self._api = api

    def _make_post_request(self, path, payload, timeout):
        req = httpx.post(self._api + path, json=payload, timeout=timeout)

        if req.status_code != 200:
            try:
                data = req.json()
                message = f"({data['message']}"
            except:
                message = ""
            raise NodeException(f"Error while doing request to api endpoint {message}")

        data = req.json()
        return data

    def load_account_data(self, account: Account):
        """
        Load the ``next_sequence`` and ``account_number`` into the account object.

        Args:
            account (Account): Account
        """

        address = account.address
        url = self._api + "/cosmos/auth/v1beta1/accounts/" + address
        req = httpx.get(url=url)
        if req.status_code != 200:
            raise NodeException("Error while doing request to api endpoint")

        data = req.json()
        sequence = int(data["account"]["sequence"])
        account_number = int(data["account"]["account_number"])

        account.next_sequence = sequence
        account.account_number = account_number

    def broadcast_transaction(self,
                              *,
                              transaction: Transaction,
                              timeout: int = 10) -> [str, int, str]:
        """
        Sign and broadcast a transaction.

        Note:
            Takes only positional arguments

        Args:
            transaction (Transaction): The transaction object
            timeout (int): Timeout

        Returns:
            hash: Transaction hash
            code: Result code
            log: Log (None if transaction successful)
        """
        path = "/cosmos/tx/v1beta1/txs"
        tx_bytes = transaction.get_tx_bytes_as_string()
        payload = {"tx_bytes": tx_bytes, "mode": "BROADCAST_MODE_SYNC"}

        data = self._make_post_request(path, payload, timeout)

        hash = data["tx_response"]["txhash"]
        code = data["tx_response"]["code"]
        log = None if code == 0 else data["tx_response"]["raw_log"]

        return {"hash": hash, "code": code, "log": log}

    def estimate_gas(self,
                              *,
                              transaction: Transaction,
                              update: bool = True,
                              multiplier: float = 1.2,
                              timeout: int = 10) ->int:
        """
        Simulate a transaction to get the estimated gas usage.

        Note:
            Takes only positional arguments

        Args:
            transaction (Transaction): The transaction object
            update (bool): Update the transaction with the estimated gas amount
            multiplier (float): Multiplier for the estimated gas when updating the transaction. Defaults to 1.2
            timeout (int): Timeout

        Returns:
            expedted_gas: Expected gas
        """
        path = "/cosmos/tx/v1beta1/simulate"
        tx_bytes = transaction.get_tx_bytes_as_string()
        payload = {"tx_bytes": tx_bytes}

        data = self._make_post_request(path, payload, timeout)

        gas_used = int(data["gas_info"]["gas_used"])

        if update:
            transaction.set_gas(int(gas_used * multiplier))

        return gas_used



from mospy import Account, Transaction
from osmosis_protobuf.osmosis.gamm.v1beta1.tx_pb2 import MsgSwapExactAmountIn
import osmosis_protobuf.osmosis.gamm.v1beta1.tx_pb2 as tx_osmosis
from mospy.clients import HTTPClient

seed_phrase = ""

account = Account(
    seed_phrase=seed_phrase,
    account_number=630264,
    next_sequence=0,
    hrp="osmo",
    protobuf="osmosis"
)
client = HTTPClient(api="https://api.osmosis.interbloc.org")

tx = Transaction( account=account, chain_id="osmosis-1", gas=10000000, protobuf="osmosis")

msg = MsgSwapExactAmountIn()
msg.sender = account.address

routes = [
    {'pool_id': "678", 'denom': "uosmo"},
    {'pool_id': '1', 'denom': 'ibc/27394FB092D2ECCD56123C74F36E4C1F926001CEADA9CA97EA622B25F41E5EB2'}
]
for route in routes:
    _route = tx_osmosis.SwapAmountInRoute()
    _route.poolId = int(route["pool_id"])
    _route.tokenOutDenom = route["denom"]
    msg.routes.append(_route)


msg.tokenIn.denom = "uatom"
msg.tokenIn.amount = str(1)
msg.tokenOutMinAmount = str(1)

tx.add_raw_msg(msg, type_url="/osmosis.gamm.v1beta1.MsgSwapExactAmountIn")
tx.set_fee( amount=25000, denom="uosmo")


print(client.broadcast_transaction(transaction=tx))



from mospy import Account, Transaction

expected_eth_tx_bytes = "Cr8BCpgBChwvY29zbW9zLmJhbmsudjFiZXRhMS5Nc2dTZW5kEngKLGV2bW9zMTV4cnYzcmN5cjBrcG54Y3k3aDloZXJsdXJrZW1tc21mbjA0ZXVqEixldm1vczE1eHJ2M3JjeXIwa3BueGN5N2g5aGVybHVya2VtbXNtZm4wNGV1ahoaCgZhZXZtb3MSEDM1MDAwMDAwMDAwMDAwMDASIlRoZSBmaXJzdCBtb3NweSBldm1vcyB0cmFuc2FjdGlvbiESfgpZCk8KKC9ldGhlcm1pbnQuY3J5cHRvLnYxLmV0aHNlY3AyNTZrMS5QdWJLZXkSIwohAhSo8BPAmQ3ByGd8vGBMgpWdkHTx/8uwzo9dP2dvm+SeEgQKAggBGAESIQobCgZhZXZtb3MSETQwMDAwMDAwMDAwMDAwMDAwEICJehpA6LslOemKjb3KTW3lsyKHOUjNXEwQwoQfuEXySvP7yugweWqSOQgIa5fh9KLCwTgE66bj5SNppaJAQpK0c53Ipw=="


class TestEthClass:
    def test_eth_transaction_creation(self):
        account = Account(
            seed_phrase="law grab theory better athlete submit awkward hawk state wedding wave monkey audit blame fury wood tag rent furnace exotic jeans drift destroy style",
            hrp="evmos",
            slip44=60,
            eth=True,
            next_sequence=1,
            account_number=2154050,
        )

        tx = Transaction(
            account=account,
            gas=2000000,
            memo="The first mospy evmos transaction!",
            chain_id="evmos_9001-2",
        )

        tx.set_fee(
            denom="aevmos",
            amount=40000000000000000
        )
        tx.add_msg(
            tx_type="transfer",
            sender=account,
            receipient=account.address,
            amount=3500000000000000,
            denom="aevmos",
        )

        assert tx.get_tx_bytes_as_string() == expected_eth_tx_bytes


from mospy import Account, Transaction
from mospy.clients import HTTPClient

SEED = ""

account = Account(
    seed_phrase=SEED,
)

client = HTTPClient(api="https://api.cosmoshub.interbloc.org")

client.load_account_data(account)

tx = Transaction(
    account=account,
    chain_id="cosmoshub-4",
    gas=250000
)

tx.add_msg('withdraw_reward', delegator=account.address,
           validator="cosmosvaloper1kgddca7qj96z0qcxr2c45z73cfl0c75p7f3s2e")
tx.add_msg('undelegate', delegator=account.address, validator="cosmosvaloper1kgddca7qj96z0qcxr2c45z73cfl0c75p7f3s2e",
           amount=500000, denom="uatom")
tx.add_msg('delegate', delegator=account.address, validator="cosmosvaloper1kgddca7qj96z0qcxr2c45z73cfl0c75p7f3s2e",
           amount=500000, denom="uatom")
tx.set_fee(
    amount=750,
    denom="uatom"
)
x = client.broadcast_transaction(transaction=tx)
print(x)



from mospy import Account


class TestAccountClass:
    seed_phrase = "law grab theory better athlete submit awkward hawk state wedding wave monkey audit blame fury wood tag rent furnace exotic jeans drift destroy style"

    def test_wallet_creation_with_seed(self):
        account = Account(seed_phrase=self.seed_phrase, hrp="osmo")

        assert account.address == "osmo1qecn0ujp4rw8hn93l9jpsxyw4fa28a52e9w9h5"
    def test_wallet_creation_with_seed_for_eth(self):
        account = Account(seed_phrase=self.seed_phrase, hrp="evmos", eth=True, slip44=60)

        assert (account.address, account.eth_address) == ("evmos15xrv3rcyr0kpnxcy7h9herlurkemmsmfn04euj", "0xa186c88f041bec199b04f5cb7c8ffc1db3bdc369")

    def test_wallet_sub_account(self):
        account = Account(seed_phrase=self.seed_phrase,
                          address_index=2,
                          next_sequence=1)
        assert account.address == "cosmos1tkv9rquxr88r7snrg42kxdj9gsnfxxg028kuh9"

        account.increase_sequence(5)

        assert account.next_sequence == 6

    def test_wallet_generation(self):
        account = Account()

        assert(len(account.seed_phrase.split())) == 24



import base64
import importlib

import grpc
from google.protobuf.json_format import MessageToDict
from mospy.Account import Account
from mospy.Transaction import Transaction


class GRPCClient:
    """
    Wrapper class to interact with a cosmos chain through their grpc endpoint

    Args:
        host (str): URL to a Cosmos api node
        port (int): Port to connect to
        ssl (bool): Whether a ssl encrypted endpoint should be used
        protobuf (str): Define which protobuf files to use. Cosmos, Evmos and Osmosis are built in and otherwise pass the raw package name (cosmospy-protobuf)

    """

    def __init__(
        self,
        *,
        host: str = "cosmoshub.strange.love",
        port: int = 9090,
        ssl: bool = False,
        protobuf="cosmos",
    ):

        _protobuf_packages = {
            "cosmos": "cosmospy_protobuf",
            "osmosis": "osmosis_protobuf",
            "evmos": "evmos_protobuf",
        }
        _protobuf_package = (_protobuf_packages[protobuf.lower()]
                             if protobuf.lower() in _protobuf_packages.keys()
                             else protobuf)
        try:
            self._cosmos_tx_service_pb2 = importlib.import_module(
                _protobuf_package +
                ".cosmos.tx.v1beta1.service_pb2")
            self._cosmos_auth_query_pb2 = importlib.import_module(
                _protobuf_package + ".cosmos.auth.v1beta1.query_pb2")
            self._cosmos_auth_query_pb2_grpc = importlib.import_module(
                _protobuf_package + ".cosmos.auth.v1beta1.query_pb2_grpc")
            self._cosmos_tx_service_pb2_grpc = importlib.import_module(
                _protobuf_package + ".cosmos.tx.v1beta1.service_pb2_grpc")

            self._BroadcastTxRequest = self._cosmos_tx_service_pb2.BroadcastTxRequest
            self._SimulateTxRequest = self._cosmos_tx_service_pb2.SimulateRequest

        except AttributeError:
            raise ImportError(
                "It seems that you are importing conflicting protobuf files. Have sou set the protobuf attribute to specify your coin? Check out the documentation for more information."
            )
        except:
            raise ImportError(
                f"Couldn't import from {_protobuf_package}. Is the package installed? "
            )

        self._host = host
        self._port = port
        self._ssl = ssl

    def _connect(self):
        if self._ssl:
            con = grpc.secure_channel(
                f"{self._host}:{self._port}",
                credentials=grpc.ssl_channel_credentials())
        else:
            con = grpc.insecure_channel(f"{self._host}:{self._port}")
        return con

    def load_account_data(self, account: Account):
        """
        Load the ``next_sequence`` and ``account_number`` into the account object.

        Args:
            account (Account): Account
        """
        con = self._connect()
        address = account.address

        query_stub = self._cosmos_auth_query_pb2_grpc.QueryStub(con)
        account_request = self._cosmos_auth_query_pb2.QueryAccountRequest(address=address)

        req = query_stub.Account(account_request)
        data = dict(MessageToDict(req.account))

        sequence = 0 if not "sequence" in data else int(data["sequence"])
        account_number = int(data["accountNumber"])

        account.next_sequence = sequence
        account.account_number = account_number
        con.close()

    def broadcast_transaction(self,
                              *,
                              transaction: Transaction,
                              timeout: int = 10) -> [str, int, str]:
        """
        Sign and broadcast a transaction.

        Note:
            Takes only positional arguments

        Args:
            transaction (Transaction): The transaction object

        Returns:
            hash: Transaction hash
            code: Result code
            log: Log (None if transaction successful)
        """
        con = self._connect()
        tx_bytes = transaction.get_tx_bytes()

        tx_request = self._BroadcastTxRequest(
            tx_bytes=tx_bytes,
            mode=2  # BROADCAST_MODE_SYNC
        )

        tx_stub = self._cosmos_tx_service_pb2_grpc.ServiceStub(con)
        tx_data = tx_stub.BroadcastTx(tx_request)

        hash = tx_data.tx_response.txhash
        code = tx_data.tx_response.code
        log = None if code == 0 else tx_data.tx_response.raw_log

        return {"hash": hash, "code": code, "log": log}

    def estimate_gas(self,
                              *,
                              transaction: Transaction,
                              update: bool = True,
                              multiplier: float = 1.2
                     ) ->int:
        """
        Simulate a transaction to get the estimated gas usage.

        Note:
            Takes only positional arguments

        Args:
            transaction (Transaction): The transaction object
            update (bool): Update the transaction with the estimated gas amount
            multiplier (float): Multiplier for the estimated gas when updating the transaction. Defaults to 1.2

        Returns:
            expedted_gas: Expected gas
        """
        con = self._connect()
        tx_bytes = transaction.get_tx_bytes()

        simulate_request = self._SimulateTxRequest(
            tx_bytes=tx_bytes,
        )

        tx_stub = self._cosmos_tx_service_pb2_grpc.ServiceStub(con)
        tx_data = tx_stub.Simulate(simulate_request)

        gas_used = tx_data.gas_info.gas_used

        if update:
            transaction.set_gas(int(gas_used * multiplier))

        return gas_used



import json

import requests
from mospy import Account, Transaction
from cosmospy_protobuf.cosmos.distribution.v1beta1.tx_pb2 import MsgSetWithdrawAddress
import httpx

SEED: str
ADDRESS: str
API = "https://rest.cosmos.directory/agoric"
RPC = "https://rpc.cosmos.directory/agoric"


account = Account(
    seed_phrase=SEED,
    slip44=564,
    next_sequence=290,
    account_number=1,
    hrp="agoric"
)
tx = Transaction(
    account=account,
    chain_id="agoric-3",
    gas=10000000
)

wmsg = MsgSetWithdrawAddress(
    delegator_address=account.address,
    withdraw_address=ADDRESS

)
tx.add_raw_msg(wmsg, type_url="/cosmos.distribution.v1beta1.MsgSetWithdrawAddress")

tx.set_fee(
    amount=0,
    denom="uist"
)

tx_bytes = tx.get_tx_bytes_as_string()

pushable_tx = json.dumps(
    {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "broadcast_tx_sync",
        # Available methods: broadcast_tx_sync, broadcast_tx_async, broadcast_tx_commit
        "params": {
            "tx": tx_bytes
        }
    }
)
r = httpx.post(RPC, data=pushable_tx)
print(r.text)


import importlib

from mnemonic import Mnemonic
from mospy.utils import privkey_to_address
from mospy.utils import privkey_to_pubkey
from mospy.utils import seed_to_private_key
from mospy.utils import privkey_to_eth_address


class Account:
    """
    The account class can be instantiated through a seed or a private key. If nothing is provided it will create a new keyring and the params to work with the cosmos chain

    Note:
        * You can't provide a ``seed_phrase`` and a``private_key``
        * A ``readble`` method behaves is the getter for a Attribute (Example: ``hrp = account.hrp``)
        * A ``writable`` method is the setter for the Attribute (Example: ``account.hrp = "cosmos"``)
        * A method can be setter and getter at the same time. The Parameters description always refers to the setter while the Returns section belongs to the getter


    Args:
        seed_phrase (str): Seed phrase to derive private keys from
        private_key (str): Private key to instantiate the Account
        next_sequence (int): Sequence which will be used for transactions signed with this Account
        account_number (int): On-chain account number
        slip44 (int): Slip44 value
        hrp (str): Address Prefix
        address_index (int): Address index to get sub accounts for seed phrases (doesn't work when using a private key)
        protobuf (str): Define which protobuf files to use. Cosmos, Evmos and Osmosis are built in and otherwise pass the raw package name (cosmospy-protobuf)
        eth (bool): Etermint compatibility mpde. If set to true the addresses and signatures will match the ethereum standard. Defaults to false to match the Cosmos standard.
    """

    address: str
    """the address of the account derived by using the slip44 param, the hrp and the address_index"""

    _RAW_DERIVATION_PATH = "m/44'/{slip44}'/0'/0/{address_index}"

    def __init__(
        self,
        seed_phrase: str = None,
        private_key: str = None,
        next_sequence: int = None,
        account_number: int = None,
        slip44: int = 118,
        hrp: str = "cosmos",
        address_index: int = 0,
        protobuf: str = "cosmos",
        eth: bool = False,
        public_key_type: str = None
    ):
        _protobuf_packages = {
            "cosmos": "cosmospy_protobuf",
            "osmosis": "osmosis_protobuf",
            "evmos": "evmos_protobuf",
        }
        _protobuf_package = (_protobuf_packages[protobuf.lower()]
                             if protobuf.lower() in _protobuf_packages.keys()
                             else protobuf)


        try:
            self.keys_pb2 = importlib.import_module(
                _protobuf_package + ".cosmos.crypto.secp256k1.keys_pb2")
        except AttributeError:
            raise ImportError(
                "It seems that you are importing conflicting protobuf files. Have sou set the protobuf attribute to specify your coin? Check out the documentation for more information."
            )
        except:
            raise ImportError(
                f"Couldn't import from {_protobuf_package}. Is the package installed? "
            )

        self._eth = eth
        self._slip44 = slip44
        self._hrp = hrp
        self._address_index = address_index
        self._next_sequence = next_sequence
        self._account_number = account_number

        # Set the public key type
        if public_key_type:
            self._public_key_type = public_key_type
        else:
            if eth:
                # Add Injectiver as exception as they use ethermint but with their own definition
                # Otherwise default to ethermint
                if hrp == "inj":
                    self._public_key_type = "/injective.crypto.v1beta1.ethsecp256k1.PubKey"
                else:
                    self._public_key_type = "/ethermint.crypto.v1.ethsecp256k1.PubKey"
            else:
                self._public_key_type = "/cosmos.crypto.secp256k1.PubKey"

        if not seed_phrase and not private_key:
            self._seed_phrase = Mnemonic(language="english").generate(
                strength=256)
            self._private_key = seed_to_private_key(self._seed_phrase,
                                                    self._derivation_path())

        elif seed_phrase and not private_key:
            self._seed_phrase = seed_phrase
            self._private_key = seed_to_private_key(seed_phrase,
                                                    self._derivation_path())

        elif private_key and not seed_phrase:
            self._seed_phrase = None
            self._private_key = bytes.fromhex(private_key)

        else:
            raise AttributeError(
                "Please set only a private key or a seed phrase. Not both!")

    def _derivation_path(self, address_index: int = None):
        adr_id = self._address_index if not address_index else address_index
        params = {"slip44": self._slip44, "address_index": adr_id}
        return self._RAW_DERIVATION_PATH.format(**params)

    @property
    def address(self) -> str:
        """
        Current address which depends on the hrp and the private key.

        Returns:
            Address
        """
        if not self._seed_phrase:
            address = privkey_to_address(self._private_key, hrp=self._hrp) if not self._eth else privkey_to_eth_address(self._private_key, hrp=self._hrp)
        else:
            sub_private_key = seed_to_private_key(
                self._seed_phrase,
                self._derivation_path(address_index=self._address_index),
            )
            address = privkey_to_address(sub_private_key, hrp=self._hrp) if not self._eth else privkey_to_eth_address(sub_private_key, hrp=self._hrp)

        return address

    @property
    def eth_address(self) -> str:
        """
        Ethereum compatible address starting with 0x. Only available if Account is initialised with eth set to True.

        Returns:
            Address
        """
        if not self._eth:
            raise TypeError("Account hasn't been initialised with the eth mode set to true.")
        if not self._seed_phrase:
            address = privkey_to_eth_address(self._private_key)
        else:
            sub_private_key = seed_to_private_key(
                self._seed_phrase,
                self._derivation_path(address_index=self._address_index),
            )
            address = privkey_to_eth_address(sub_private_key)

        return address

    @property
    def seed_phrase(self) -> str:
        """
        Current Seed Phrase

        Returns:
            Seed Phrase
        """
        return self._seed_phrase

    @property
    def private_key(self) -> bytes:
        """
        Current private key which depends on the slip 44 param and the address index if the account is instantiated through a seed.

        Returns:
            Private Key
        """
        if self._seed_phrase:
            private_key = seed_to_private_key(
                self._seed_phrase,
                self._derivation_path(address_index=self._address_index),
            )
            return private_key
        else:
            return self._private_key

    @property
    def public_key(self) -> str:
        """
        Current public key which depends on the slip 44 param and the address index if the account is instantiated through a seed.

        Returns:
            Public Key
        """
        pubkey_bytes = privkey_to_pubkey(self.private_key)
        _pubkey = self.keys_pb2.PubKey()
        _pubkey.key = pubkey_bytes
        return _pubkey

    @property
    def account_number(self) -> int:
        """
        On-chain account number which will be assigned when the address receives coins for the first time.

        Returns:
            Account number
        """
        return self._account_number

    @account_number.setter
    def account_number(self, account_number: int):
        self._account_number = account_number

    @property
    def next_sequence(self) -> int:
        """
        Sequence which will be used for transactions signed with this Account.

        Returns:
            Next Sequence
        """
        return self._next_sequence

    @next_sequence.setter
    def next_sequence(self, next_sequence):
        self._next_sequence = next_sequence

    def increase_sequence(self, change: int = 1) -> None:
        """
        Increase the sequence by ``change``

        Args:
            change (int): Value to increase the sequence
        """
        self._next_sequence += change

    @property
    def address_index(self):
        """
        Change the address index to use a sub account. This works only if a seed has been used to instantiate the Account.


        Returns:
            Address Index
        """
        return self._address_index

    @address_index.setter
    def address_index(self, address_index: int) -> None:

        if self._seed_phrase:
            self._DEFAULT_ADDRESS_INDEX = address_index
        else:
            raise ValueError(
                "Can't the change the address index without provided seed")

    @property
    def hrp(self) -> str:
        """
        Current address prefix used by the Account.

        Returns:
            Address Prefix (hrp)
        """
        return self._hrp

    @hrp.setter
    def hrp(self, hrp: str) -> None:
        self._hrp = hrp

    @property
    def slip44(self) -> int:
        """
        Set the Slip44 value. Cosmos defaults to 118

        Returns:
            Slip44

        """
        return self._slip44

    @slip44.setter
    def slip44(self, slip44: int) -> None:
        self._slip44 = slip44

    @property
    def eth(self) -> bool:
        """
        Change the eth compatibility mode. If you want to use Evmos you will need to set eth to true. Otherwise it defaults to False

        Returns:
            eth
          """
        return self._eth

    @eth.setter
    def eth(self, eth: bool):

        self._eth = eth

    @property
    def public_key_type(self):
        """
        Get the current pyblic key type

        Returns:
            public_key_type
          """
        return self._public_key_type
    @public_key_type.setter
    def public_key_type(self, public_key_type):
        self._public_key_type = public_key_type



import hashlib

import ecdsa
import hdwallets
import bech32
from mnemonic import Mnemonic
from sha3 import keccak_256
import binascii


def seed_to_private_key(seed, derivation_path, passphrase: str = ""):
    seed_bytes = Mnemonic.to_seed(seed, passphrase=passphrase)
    hd_wallet = hdwallets.BIP32.from_seed(seed_bytes)
    # This can raise a `hdwallets.BIP32DerivationError` (which we alias so
    # that the same exception type is also in the `cosmospy` namespace).
    derived_privkey = hd_wallet.get_privkey_from_path(derivation_path)

    return derived_privkey


def privkey_to_pubkey(privkey: bytes, raw: bool = False) -> bytes:
    privkey_obj = ecdsa.SigningKey.from_string(privkey, curve=ecdsa.SECP256k1)
    pubkey_obj = privkey_obj.get_verifying_key()
    return pubkey_obj.to_string("raw") if raw else pubkey_obj.to_string("compressed")


def pubkey_to_address(pubkey: bytes, *, hrp: str) -> str:
    s = hashlib.new("sha256", pubkey).digest()
    r = hashlib.new("ripemd160", s).digest()
    five_bit_r = bech32.convertbits(r, 8, 5)
    assert five_bit_r is not None, "Unsuccessful bech32.convertbits call"
    return bech32.bech32_encode(hrp, five_bit_r)


def pubkey_to_eth_address(pubkey: bytes, *, hrp: str = None) -> str:
    k = keccak_256(pubkey).digest()[-20:].hex()
    if not hrp:
        return "0x" + k

    array = binascii.unhexlify(k)
    words = [x for x in array]
    bech32_words = bech32.convertbits(words, 8, 5)
    bech32_address = bech32.bech32_encode(hrp, bech32_words)
    return bech32_address


def privkey_to_eth_address(privkey: bytes, *, hrp: str = None) -> str:
    pubkey = privkey_to_pubkey(privkey, raw=True)
    return pubkey_to_eth_address(pubkey, hrp=hrp)


def privkey_to_address(privkey: bytes, *, hrp: str) -> str:
    pubkey = privkey_to_pubkey(privkey)
    return pubkey_to_address(pubkey, hrp=hrp)



from cosmospy_protobuf.cosmos.base.v1beta1.coin_pb2 import Coin
from mospy import Account
from mospy import Transaction

expected_tx_bytes = "CpABCo0BChwvY29zbW9zLmJhbmsudjFiZXRhMS5Nc2dTZW5kEm0KLWNvc21vczFxZWNuMHVqcDRydzhobjkzbDlqcHN4eXc0ZmEyOGE1MjM3YTRweBItY29zbW9zMXRrdjlycXV4cjg4cjdzbnJnNDJreGRqOWdzbmZ4eGcwMjhrdWg5Gg0KBXVhdG9tEgQxMDAwEmYKUApGCh8vY29zbW9zLmNyeXB0by5zZWNwMjU2azEuUHViS2V5EiMKIQIkQVvG9OBetDe7bYUgl2vwgbJFxmmGuquytVSEwhQ0uBIECgIIARgBEhIKDQoFdWF0b20SBDEwMDAQ6AcaQJqww3jDNgn4UMDpaFq34xPbdwTAsn4VnRvZ1rjYGCMEa6fDKnu9T5xlQV5IEpCDeMzmNBEhTo9QtcOIzjjPzes="

class TestTransactionClass:
    seed_phrase = "law grab theory better athlete submit awkward hawk state wedding wave monkey audit blame fury wood tag rent furnace exotic jeans drift destroy style"

    def test_transaction_creation(self):
        account = Account(
            seed_phrase=
            "law grab theory better athlete submit awkward hawk state wedding wave monkey audit blame fury wood tag rent furnace exotic jeans drift destroy style",
            account_number=1,
            next_sequence=1,
        )

        fee = Coin(denom="uatom", amount="1000")

        tx = Transaction(
            account=account,
            fee=fee,
            gas=1000,
        )

        tx.add_msg(
            tx_type="transfer",
            sender=account,
            receipient="cosmos1tkv9rquxr88r7snrg42kxdj9gsnfxxg028kuh9",
            amount=1000,
            denom="uatom",
        )


        tx_bytes = tx.get_tx_bytes_as_string()
        assert tx_bytes == expected_tx_bytes




import importlib

class Transaction:
    name = "delegate"  # Name to reference this transaction type
    _type_url = "/cosmos.staking.v1beta1.MsgDelegate"

    def __init__(
        self,
        protobuf_package: str,
        delegator: str,
        validator: str,
        amount: int,
        denom: str,
    ):
        try:
            self._coin_pb2 = importlib.import_module(
                protobuf_package + ".cosmos.base.v1beta1.coin_pb2")
            self._tx_pb2 = importlib.import_module(
                protobuf_package + ".cosmos.staking.v1beta1.tx_pb2")
        except:
            raise ImportError(
                f"Couldn't import from {protobuf_package}. Is the package installed?"
            )

        _tx_coin = self._coin_pb2.Coin()
        _tx_coin.denom = denom
        _tx_coin.amount = str(amount)

        self._amount = _tx_coin
        self._delegator_address = delegator
        self._validator = validator

    def format(self) -> (str, object):
        msg = self._tx_pb2.MsgUndelegate(
            delegator_address=self._delegator_address,
            validator_address=self._validator,
            amount=self._amount
        )

        return (self._type_url, msg)



import base64
import hashlib
import importlib

import ecdsa
from sha3 import keccak_256
from google.protobuf import any_pb2 as any
from mospy._transactions import ALL_TRANSACTION_HELPERS
from mospy.Account import Account

built_in_transactions = {}
for transaction_adapter in ALL_TRANSACTION_HELPERS:
    module = importlib.import_module("mospy._transactions." +
                                     transaction_adapter)
    adapter = getattr(module, "Transaction")
    built_in_transactions[adapter.name] = adapter


class Transaction:
    """Class to create and sign a transaction
    Args:
        account (Account): Account object to sign this transaction
        gas (int): Gas unit for this transaction
        fee (coin): The fee to pay for this transaction (This can also be added later through the ``set_fee`` method)
        memo (str): Memo
        chain_id (str): Chain-Id "cosmoshub-4",
        protobuf (str): Define which protobuf files to use. Cosmos, Evmos and Osmosis are built in and otherwise pass the raw package name (cosmospy-protobuf)


    """

    def __init__(
        self,
        *,
        account: Account,
        gas: int,
        fee: object = None,
        memo: str = "",
        chain_id: str = "cosmoshub-4",
        protobuf: str = "cosmos",
        feegrant: str = ""
    ) -> None:

        _protobuf_packages = {
            "cosmos": "cosmospy_protobuf",
            "osmosis": "osmosis_protobuf",
            "evmos": "evmos_protobuf",
        }
        self._protobuf_package = (_protobuf_packages[protobuf.lower()]
                                  if protobuf.lower()
                                  in _protobuf_packages.keys() else protobuf)
        try:
            self.coin_pb2 = importlib.import_module(
                self._protobuf_package + ".cosmos.base.v1beta1.coin_pb2")
            self.tx_pb2 = importlib.import_module(self._protobuf_package +
                                                  ".cosmos.tx.v1beta1.tx_pb2")
        except:
            raise ImportError(
                f"Couldn't import from {self._protobuf_package}. Is the package installed?"
            )

        if fee and not isinstance(fee, (self.coin_pb2.Coin)):
            raise ValueError("The fee is not a valid.")

        self._account = account
        self._fee = fee
        self._gas = gas
        self._chain_id = chain_id
        self._tx_body = self.tx_pb2.TxBody()
        self._tx_body.memo = memo
        self._tx_raw = self.tx_pb2.TxRaw()
        self._feegrant = feegrant

    def add_msg(self, tx_type: str, **kwargs) -> None:
        """
        Add a pre-defined message to the tx body.

        Args:
            tx_type (str): Transaction type to match the transaction with the pre-defined ones
            **kwargs: Depending on the transaction type
        """
        msg_data = built_in_transactions[tx_type](
            protobuf_package=self._protobuf_package, **kwargs).format()
        self.add_raw_msg(msg_data[1], type_url=msg_data[0])

    def add_raw_msg(self, unpacked_msg, type_url: str) -> None:
        """
        Add a message to the tx body manually.

        Args:
            unpacked_msg: Transaction data
            type_url: Type url for the transaction
        """
        msg_any = any.Any()
        msg_any.Pack(unpacked_msg)
        msg_any.type_url = type_url
        self._tx_body.messages.append(msg_any)

    def set_fee(self, amount: int, denom: str = "uatom"):
        """
        Set the fee manually

        Args:
            amount: Amount
            denom: Denom
        """
        self._fee = self.coin_pb2.Coin(amount=str(amount), denom=denom)

    def set_gas(self, gas: int):
        """
        Update the wanted gas for the transaction

        Args:
            gas: Gas
        """
        self._gas = gas

    def get_tx_bytes(self) -> bytes:
        """Sign the transaction and get the tx bytes which can be used to broadcast the transaction to the network.

        To broadcast the transaction through the REST endpoint use the ``get_tx_bytes_as_string()`` method instead

        Returns:
            tx_bytes (bytes): Transaction bytes
        """
        self._tx_raw = self.tx_pb2.TxRaw()
        self._tx_raw.body_bytes = self._tx_body.SerializeToString()
        self._tx_raw.auth_info_bytes = self._get_auth_info().SerializeToString(
        )
        self._tx_raw.signatures.append(self._get_signatures())
        raw_tx = self._tx_raw.SerializeToString()
        tx_bytes = bytes(raw_tx)
        return tx_bytes

    def get_tx_bytes_as_string(self) -> str:
        """Sign the transaction and get the base64 encoded tx bytes which can be used to broadcast the transaction to the network.

        Returns:
            tx_bytes (str): Transaction bytes
        """
        tx_bytes = self.get_tx_bytes()
        tx_b64 = base64.b64encode(tx_bytes).decode("utf-8")
        return tx_b64

    def _get_signatures(self):
        privkey = ecdsa.SigningKey.from_string(self._account.private_key,
                                               curve=ecdsa.SECP256k1)

        # Cosmos uses sha256 as main hashing function while ethereum uses keccak256
        hashfunc = hashlib.sha256 if not self._account.eth else keccak_256

        signature_compact = privkey.sign_deterministic(
            self._get_sign_doc().SerializeToString(),
            hashfunc=hashfunc,
            sigencode=ecdsa.util.sigencode_string_canonize,
        )
        return signature_compact

    def _get_sign_doc(self):
        sign_doc = self.tx_pb2.SignDoc()
        sign_doc.body_bytes = self._tx_body.SerializeToString()
        sign_doc.auth_info_bytes = self._get_auth_info().SerializeToString()
        sign_doc.chain_id = self._chain_id
        sign_doc.account_number = self._account.account_number
        return sign_doc

    def _get_auth_info(self):
        _auth_info = self.tx_pb2.AuthInfo()
        _auth_info.signer_infos.append(self._get_signer_infos())
        _auth_info.fee.gas_limit = self._gas
        _auth_info.fee.amount.append(self._fee)
        if self._feegrant:
            _auth_info.fee.granter = self._feegrant
        return _auth_info

    def _get_signer_infos(self):
        signer_infos = self.tx_pb2.SignerInfo()
        signer_infos.sequence = self._account.next_sequence
        signer_infos.public_key.Pack(self._account.public_key)
        signer_infos.public_key.type_url = self._account.public_key_type
        signer_infos.mode_info.single.mode = 1
        return signer_infos



from cosmospy_protobuf.cosmos.base.v1beta1.coin_pb2 import Coin
from mospy import Account
from mospy import Transaction
from mospy.clients import GRPCClient


class TestGRPCClientClass:
    seed_phrase = "law grab theory better athlete submit awkward hawk state wedding wave monkey audit blame fury wood tag rent furnace exotic jeans drift destroy style"

    def test_account_data_loading(self):
        account = Account(seed_phrase=self.seed_phrase)

        client = GRPCClient(host="grpc-cosmoshub.whispernode.com",
                            port=443,
                            ssl=True)

        client.load_account_data(account)

        assert account.account_number == 1257460

        assert account.next_sequence >= 0

    def test_transaction_submitting(self):
        account = Account(
            seed_phrase=self.seed_phrase,
            account_number=1,
            next_sequence=0,
        )

        client = GRPCClient(host="grpc-cosmoshub.whispernode.com",
                            port=443,
                            ssl=True)

        fee = Coin(denom="uatom", amount="1000")

        tx = Transaction(
            account=account,
            fee=fee,
            gas=10000000000,
        )

        tx.add_msg(
            tx_type="transfer",
            sender=account,
            receipient="cosmos1tkv9rquxr88r7snrg42kxdj9gsnfxxg028kuh9",
            amount=1000,
            denom="uatom",
        )

        tx_data = client.broadcast_transaction(transaction=tx)

        assert (
            tx_data["hash"] ==
            "54B845AEB1523803D4EAF2330AE5759A83458CB5F0211159D04CC257428503C4")


        client.load_account_data(account=account)

        gas_used = client.estimate_gas(
            transaction=tx,
            update=False,
        )

        assert gas_used > 0



def __list_all_transaction_helpers():
    from os.path import dirname, basename, isfile
    import glob

    mod_paths = glob.glob(dirname(__file__) + "/*.py")
    all_modules = [basename(f)[:-3] for f in mod_paths if isfile(f)
                   and f.endswith(".py")
                   and not f.endswith('__init__.py')]

    return all_modules


ALL_TRANSACTION_HELPERS = sorted(__list_all_transaction_helpers())



from cosmospy_protobuf.ibc.applications.transfer.v1.tx_pb2 import MsgTransfer
from cosmospy_protobuf.cosmos.base.v1beta1.coin_pb2 import Coin
from cosmospy_protobuf.ibc.core.client.v1.client_pb2 import Height # Optional, see comments below

from src.mospy import Transaction, Account
from src.mospy.clients import HTTPClient

client = HTTPClient()
account = Account(
    seed_phrase="SEED HERE",
    account_number=1234,
    next_sequence=1234
)
tx = Transaction(
    account=account,
    gas=10000000  # This need to be adapted
)

asset = Coin(
    amount="1000000",
    denom="uatom"
)

# Use either a timeout_tiemstamp or timeout_height for your transfer If you want to disable it you can set both to
# zero. But keep in mind that you need one of the two to get the transfer working More information about timeouts:
# https://ibc.cosmos.network/main/ibc/overview.html#receipts-and-timeouts

timeout_height = Height(
    revision_number=4,
    revision_height=1234
)
tmsg = MsgTransfer(
    source_port="transfer",
    source_channel="channel-1",
    token=asset,
    sender=account.address,
    receiver="",
    timeout_timestamp=0,
    timeout_height=timeout_height,

)

tx.set_fee(
    amount=2000,
    denom="uatom"
)


tx.add_raw_msg(tmsg, type_url="/ibc.applications.transfer.v1.MsgTransfer")

client.broadcast_transaction(transaction=tx)


import importlib

from mospy.Account import Account


class Transaction:
    name = "transfer"  # Name to reference this transaction type
    _type_url = "/cosmos.bank.v1beta1.MsgSend"

    def __init__(
        self,
        protobuf_package: str,
        sender: Account,
        receipient: str,
        amount: int,
        denom: str,
    ):
        try:
            self._coin_pb2 = importlib.import_module(
                protobuf_package + ".cosmos.base.v1beta1.coin_pb2")
            self._tx_pb2 = importlib.import_module(
                protobuf_package + ".cosmos.bank.v1beta1.tx_pb2")
        except:
            raise ImportError(
                f"Couldn't import from {protobuf_package}. Is the package installed?"
            )

        _tx_coin = self._coin_pb2.Coin()
        _tx_coin.denom = denom
        _tx_coin.amount = str(amount)
        self._amount = _tx_coin
        self._sender = sender.address
        self._receipient = receipient

    def format(self) -> (str, object):
        msg = self._tx_pb2.MsgSend(
            from_address=self._sender,
            to_address=self._receipient,
        )
        msg.amount.append(self._amount)

        return (self._type_url, msg)
